#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "UTFChar.h"
#include "UTFString.h"


//Constructors
UTFString::UTFString()
:	m_pBuffer(NULL),
	m_length(0)
{
}

UTFString::UTFString(const char* rhs)
:	m_pBuffer(NULL),
	m_length(0)
{
	CopyString(rhs);
}

UTFString::UTFString(const UTFChar& rhs)
:	m_pBuffer(NULL),
	m_length(0)
{
	CopyString(rhs);
}

UTFString::UTFString(const UTFString& rhs)
:	m_pBuffer(NULL),
	m_length(0)
{
	CopyString(rhs.m_pBuffer);
}
	
//Destructors
UTFString::~UTFString()
{
	ClearBuffer();
}
	
//Operator Overloads
UTFString& UTFString::operator=(const char* rhs)
{
	BAutolock(this);
	ClearBuffer();
	CopyString(rhs);
	return *this;
}

UTFString& UTFString::operator=(const UTFString& rhs)
{
	BAutolock(this);
	
	if (rhs.m_pBuffer == m_pBuffer)
	{
		return *this;
	}
		
	ClearBuffer();
	CopyString(rhs.m_pBuffer);
	return *this;
}

void UTFString::Insert(uint32 bindex, uint32 eindex, const UTFString& instr)
{

	BAutolock(this);

	
	if (!m_pBuffer)
	{
		operator=(instr);
	}
	else
	{
		UTFString lhs;
		UTFString rhs;
		UTFString rhs2;
		UTFString rhs3;
		
		Split(bindex, lhs, rhs);
		rhs.Split(eindex-bindex, rhs3, rhs2);
	
		operator=((lhs + instr + rhs2));
	}
	
	CalcLength();

	
}

UTFString UTFString::Remove(uint32 bindex, uint32 eindex)
{
	UTFString tstr;
	UTFString tstr2;
	UTFString tstr3;
	UTFString tstr4;
	
	BAutolock(this);

	
	if (m_pBuffer)
	{
		assert(eindex > bindex);
	
		Split(eindex, tstr, tstr2);
		tstr.Split(bindex, tstr4, tstr3);
	
		operator=((tstr4 + tstr2));
	
		CalcLength();
	}

	
		
	return tstr2;
}

UTFString UTFString::Left(uint32 idx)
{
	assert(idx > 0);
	assert(idx <= m_length);

	UTFString lhs;
	UTFString rhs;
	
	Split(idx - 1, lhs, rhs);
		
	return lhs;
}

UTFString UTFString::Right(uint32 idx)
{
	assert(idx > 0);
	assert(idx < m_length);

	idx = m_length - idx;

	UTFString lhs;
	UTFString rhs;
	
	Split(idx, lhs, rhs);
	
	return rhs;
}

UTFString UTFString::Mid(uint32 bidx, uint32 eidx)
{
	eidx += bidx;
	
	assert(bidx >= 0);
	assert(bidx < m_length);

	assert(eidx >= 0);
	if (eidx >= m_length)
		eidx = m_length-1;
	
	eidx -= bidx;
	
	assert(bidx <= eidx);
	
	UTFString lhs;
	UTFString mid;
	UTFString rhs;
		
	Split(bidx, lhs, rhs);
	rhs.Split(eidx, mid, lhs);
	
	return mid;
}

void UTFString::Split(uint32 index, UTFString& lhs, UTFString& rhs)
{
	if (m_pBuffer)
	{
		assert(index >= 0);
		assert(index <= m_length);
	
		if (0 == index)
		{
			rhs = *this;
		}
//		else if ((m_length-1) == index)
//		{
//			lhs = *this;
//		}
		else
		{
			uint32 idx = 0;
			uint32 byteidx = 0;

			BAutolock(this);

					
			for (idx = 0; idx < index; idx++)
			{
				byteidx += UTF8CHARBYTELEN(m_pBuffer[byteidx]);
			}
		
			char t;
		
			t = m_pBuffer[byteidx];
		
			m_pBuffer[byteidx] = '\0';
			lhs = m_pBuffer;
		
			m_pBuffer[byteidx] = t;
			rhs = (m_pBuffer + byteidx);
			
			
		}
	}
}

bool UTFString::Split(const UTFString& delim, UTFString& lhs, UTFString& rhs)
{
	bool retval = false;
	
	assert(0 < delim.Length());
	
	if (m_pBuffer)
	{
		BAutolock(this);

		int32 subpos;
	
		subpos = InStr(delim);

		if (subpos > -1)
		{
			UTFString mid;
			UTFString mid2;
			UTFString trhs;
			
			Split(subpos, lhs, mid);
			
			if (delim.Length() <= mid.Length())
				mid.Split(delim.Length() , mid2, trhs);
			
			rhs = trhs;
			retval = true;
		}
		else
		{
			lhs = *this;
		}
		
	}	
	
	return retval;
}

int32 UTFString::InStr(const UTFString& substr)
{
	char* subpos = NULL;
	int32 cpos = -1;
	
	if (!m_pBuffer)
		return -1;
		
	assert(0 < substr.Length());

	BAutolock(this);

	subpos = strstr(m_pBuffer, substr.m_pBuffer);
	
	if (subpos)
	{
		char* byteidx = m_pBuffer;
		cpos = 0;
		
		while(subpos != byteidx && *byteidx)
		{
			cpos++;
			byteidx += UTF8CHARBYTELEN(*byteidx);
		}
		
	}

	
	
	return(cpos);

}

int32 UTFString::InStr(int spos, const UTFString& substr)
{
	char* subpos = NULL;
	int32 cpos = -1;
	
	if (!m_pBuffer)
		return -1;
		
	assert(0 < substr.Length());
	assert(0 <= spos);
	assert(spos < m_length);

	BAutolock(this);

	
	char* phead;
	phead = m_pBuffer;

	for (int32 idx = 0; idx < spos; idx++)
	{
		phead += UTF8CHARBYTELEN(*phead);
	}

	subpos = strstr(phead, substr.m_pBuffer);
	
	if (subpos)
	{
		char* byteidx = phead;
		cpos = spos;
		
		while(subpos != byteidx && *byteidx)
		{
			cpos++;
			byteidx += UTF8CHARBYTELEN(*byteidx);
		}
		
	}

	
	return(cpos);
}

void UTFString::Replace(const UTFString& substr, const UTFString& repstr)
{
	int32 subpos = 0;
	
	UTFString lhs;
	UTFString rhs;
	
	assert(0 < substr.Length());
	assert(0 < repstr.Length());

	subpos = InStr(substr);
	
	while(subpos > -1)
	{
		
		UTFString t;
		
		lhs = Left(subpos);
		BAutolock(this);

		rhs = Right(m_length - (subpos + substr.Length()));
		//Split(substr, lhs, rhs);

		t = lhs + repstr + rhs;

		operator=(t);
		
		subpos = InStr(subpos + substr.Length(), substr);
		
	}
}

void UTFString::ToUpper()
{
	uint32 idx = 0;
	uint32 byteidx = 0;
		
	BAutolock(this);

	for (idx = 0; idx < m_length; idx++)
	{
		if (m_pBuffer[byteidx] < 0x80)
			m_pBuffer[byteidx] = toupper(m_pBuffer[byteidx]);
			
		byteidx += UTF8CHARBYTELEN(m_pBuffer[byteidx]);
		
	}
	
}

void UTFString::ToLower()
{
	uint32 idx = 0;
	uint32 byteidx = 0;
		
	BAutolock(this);

	for (idx = 0; idx < m_length; idx++)
	{
		if (m_pBuffer[byteidx] < 0x80)
			m_pBuffer[byteidx] = tolower(m_pBuffer[byteidx]);

		byteidx += UTF8CHARBYTELEN(m_pBuffer[byteidx]);
	}
	
}

UTFChar UTFString::GetChar(uint32 index)
{
	UTFChar tchar;
	uint32 idx = 0;
	uint32 byteidx = 0;
		
	if (!m_pBuffer)
		return tchar;
		
	BAutolock(this);

	for (idx = 0; idx < index; idx++)
	{
		byteidx += UTF8CHARBYTELEN(m_pBuffer[byteidx]);
	}
	
	tchar = &m_pBuffer[byteidx];
	
	
	return tchar;
}

void UTFString::SetChar(const UTFChar& inchar, uint32 index)
{
	Insert(index, index+1, inchar);
}

void UTFString::CalcLength()
{
	m_length = 0;
	char* pHead = NULL;
	
	if (!m_pBuffer)
		return;
		
	BAutolock(this);

	pHead = m_pBuffer;
	
	while(*pHead)
	{
		pHead += UTF8CHARBYTELEN(*pHead);
		m_length++;
	}
	
}

void UTFString::ClearBuffer()
{
	if (m_pBuffer)
	{
		BAutolock(this);

		free(m_pBuffer);
		m_pBuffer = NULL;
		
	}
}

void UTFString::StrCat(const char* rhs)
{
	char* d;
	int32 llen;
	int32 rlen;
	int32 len;
	
	BAutolock(this);

	llen = strlen(m_pBuffer);
	rlen = strlen(rhs);
	
	len = llen + rlen + 1;

	d = new char[len];
	memset(d, '\0', len);

	char* chead;
	chead = d;

	strcpy(chead, m_pBuffer);
	strcpy((chead+llen), rhs);
	
	ClearBuffer();
	
	m_pBuffer = d;
	
}

void UTFString::CopyString(const char* tstr)
{
	if (tstr)
	{
		BAutolock(this);

		m_pBuffer = (char*)malloc(strlen(tstr) + 1);
		memset(m_pBuffer, '\0', strlen(tstr) + 1);
		strcpy(m_pBuffer, tstr);
		
	}
	
	CalcLength();
}

UTFString::operator const char*() const
{
	return m_pBuffer;
}

bool UTFString::operator==(const char* rhs) const
{
	if (m_pBuffer && rhs)
	{
		return (strcmp(m_pBuffer, rhs))?false:true;
	}
	else
	{
		return false;
	}
}

bool UTFString::operator==(const UTFString& rhs) const
{
	if (m_pBuffer)
	{
		return (strcmp(m_pBuffer, rhs))?false:true;
	}
	else
	{
		return false;
	}
}

bool UTFString::operator!=(const char* rhs)
{
	return !operator==(rhs);
}

bool UTFString::operator!=(const UTFString& rhs)
{
	return !operator==(rhs);
}
			
UTFString& UTFString::operator+=(const char* rhs)
{
	if (!m_pBuffer)
	{
		operator=(rhs);
	}
	else
	{
		if (rhs)
		{
			StrCat(rhs);
			CalcLength();
		}
	}
	
	return *this;
}

UTFString& UTFString::operator+=(const UTFString& rhs)
{
	if (!m_pBuffer)
	
	{
		operator=(rhs);
	}
	else
	{
		if (rhs.m_pBuffer)
		{
			StrCat(rhs.m_pBuffer);
			CalcLength();
		}
	}
	
	return *this;
}
	
UTFString operator+(const UTFString& lhs, const UTFString& rhs)
{
	UTFString t(lhs);
	
	t+=rhs;
	
	return t;
}
	
uint32 UTFString::Length() const
{
	return m_length;
}

uint32 UTFString::ByteLength() const
{
	if (m_pBuffer)
	{
		return strlen(m_pBuffer);
	}
	else
	{
		return 0;
	}
}

